import time
from typing import Any, Dict, Iterator, Optional

from finetuner.client import FinetunerV1Client
from finetuner.console import console, print_examples, print_metrics
from finetuner.constants import (
    ARTIFACT_ID,
    ARTIFACTS_DIR,
    CREATED,
    FAILED,
    STARTED,
    STATUS,
    SYNTHESIS_TASK,
    TRAIN_DATA,
    TRAINING_TASK,
)
from finetuner.excepts import RunFailedError, RunInProgressError, RunPreparingError
from finetuner.hubble import download_artifact


class Run:
    """Class for a run.

    :param client: Client object for sending api requests.
    :param name: Name of the run.
    :param experiment_name: Name of the experiment.
    :param config: Configuration for the run.
    :param created_at: Creation time of the run.
    :param description: Optional description of the run.
    :param train_data: The name of the `DocumentArray` created if this run is a data
        synthesis job.
    """

    def __init__(
        self,
        client: FinetunerV1Client,
        name: str,
        experiment_name: str,
        config: dict,
        created_at: str,
        description: str = '',
        task: str = TRAINING_TASK,
        train_data: Optional[str] = None,
    ):
        self._client = client
        self._name = name
        self._experiment_name = experiment_name
        self._config = config
        self._created_at = created_at
        self._description = description
        self._run = self._get_run()
        self.task = task
        self._train_data = train_data

    @property
    def name(self) -> str:
        """Get the name of the :class:`Run`."""
        return self._name

    @property
    def config(self) -> dict:
        """Get the config of the :class:`Run`."""
        return self._config

    @property
    def train_data(self) -> str:
        """Get the data generated by the :class:`Run` In the case that it was a
        Synthesis job, if it is a training job, a ValueError is thrown.
        """
        if self.task != SYNTHESIS_TASK:
            raise ValueError(f'{self.task} run does not produce data.')
        else:
            self._check_run_status_finished()
            if self._train_data:
                return self._train_data
            else:
                run = self._get_run()
                try:
                    train_data = run[TRAIN_DATA]
                except KeyError:
                    raise ValueError(f'run {self.name} has no train_data.')
                self._train_data = train_data
                return train_data

    def _get_run(self) -> dict:
        """Get Run object as dict."""
        return self._client.get_run(
            experiment_name=self._experiment_name, run_name=self._name
        )

    def status(self) -> dict:
        """Get :class:`Run` status.

        :returns: A dict representing the :class:`Run` status.
        """
        return self._client.get_run_status(
            experiment_name=self._experiment_name, run_name=self._name
        )

    def logs(self) -> str:
        """Check the :class:`Run` logs.

        :returns: A string dump of the run logs.
        """
        self._check_run_status_started()
        return self._client.get_run_logs(
            experiment_name=self._experiment_name, run_name=self._name
        )

    def stream_logs(self, interval: int = 5) -> Iterator[str]:
        """Stream the :class:`Run` logs lively.

        :param interval: The time interval to sync the status of finetuner `Run`.
        :yield: An iterators keep stream the logs from server.
        """
        status = self.status()[STATUS]
        msg_template = (
            'Preparing to run, logs will be ready to pull when '
            '`status` is `STARTED`. Current status is `%s`'
        )
        with console.status(msg_template % status, spinner="dots") as rich_status:
            while status == CREATED:
                time.sleep(interval)
                status = self.status()[STATUS]
                rich_status.update(msg_template % status)

        return self._client.stream_run_logs(
            experiment_name=self._experiment_name, run_name=self._name
        )

    def metrics(self) -> Dict[str, Dict[str, float]]:
        """Get the evaluation metrics of the :class:`Run`.

        :return: dictionary with evaluation metrics before and after fine-tuning.
        """
        self._check_run_status_finished()
        return self._client.get_run_metrics(
            experiment_name=self._experiment_name, run_name=self._name
        )

    def display_metrics(self):
        """
        Prints a table of retrieval metrics before and after fine-tuning
        """
        metrics = self.metrics()
        for stage in metrics:
            print_metrics(stage, metrics[stage])

    def example_results(self) -> Dict[str, Any]:
        """Get the results of example queries from the evaluation data of the
        :class:`Run`.

        :return: dictionary with results before and after fine-tuning.
        """
        self._check_run_status_finished()
        return self._client.get_run_examples(
            experiment_name=self._experiment_name, run_name=self._name
        )

    def display_examples(self, k: int = 5):
        """
        Prints a table of results of example queries before and after fine-tuning.

        :param k: maximal number of results per query to display
        """
        example_results = self.example_results()
        for stage in example_results:
            print_examples(stage, example_results[stage], k=k)

    def _check_run_status_finished(self):
        status = self.status()[STATUS]
        if status in [CREATED, STARTED]:
            raise RunInProgressError(
                'The run needs to be finished in order to save the artifact.'
            )
        if status == FAILED:
            raise RunFailedError(
                'The run failed, please check the `logs` for detailed information.'
            )

    def _check_run_status_started(self):
        status = self.status()[STATUS]
        if status == CREATED:
            raise RunPreparingError(
                'Preparing to run, logs will be ready to pull when '
                '`status` is `STARTED`.'
            )

    def save_artifact(self, directory: str = ARTIFACTS_DIR) -> str:
        """Save artifact if the :class:`Run` is finished.

        :param directory: Directory where the artifact will be stored.
        :returns: A string object that indicates the download path.
        """
        self._check_run_status_finished()
        return download_artifact(
            client=self._client,
            artifact_id=self._run[ARTIFACT_ID],
            run_name=self._name,
            directory=directory,
        )

    @property
    def artifact_id(self):
        """Get artifact id of the :class:`Run`.

        An artifact in finetuner contains fine-tuned model and its metadata.
        Such as preprocessing function, collate function. This id could be useful
        if you want to directly pull the artifact from the cloud storage, such as
        using `FinetunerExecutor`.

        :return: Artifact id as string object.
        """
        self._check_run_status_finished()
        return self._run[ARTIFACT_ID]



import os
from typing import Dict, Optional, Tuple, Union

from finetuner import DocumentArray
from finetuner.constants import ARTIFACTS_DIR, DA_PREFIX


def push_docarray(
    data: Union[None, str, DocumentArray],
    name: str,
    ids2names: Optional[Dict[int, str]] = None,
) -> Optional[str]:
    """Upload a DocumentArray to Jina AI Cloud and return its name."""
    if isinstance(data, DocumentArray):
        _id = id(data)  # get the reference id
        if ids2names is not None and _id in ids2names:
            return ids2names[_id]
        print(f'Pushing a DocumentArray to Jina AI Cloud under the name {name} ...')
        data.push(name=name, show_progress=True, public=False)
        if ids2names is not None:
            ids2names[id(data)] = name
        return name
    return data


def push_training_data(
    experiment_name: str,
    run_name: str,
    train_data: Union[str, DocumentArray],
    eval_data: Union[None, str, DocumentArray] = None,
    query_data: Union[None, str, DocumentArray] = None,
    index_data: Union[None, str, DocumentArray] = None,
) -> Tuple[Optional[str], ...]:
    """Upload data to Jina AI Cloud and returns their names.

    Uploads all data needed for fine-tuning - training data,
    evaluation data and query/index data for `EvaluationCallback`.

    Data is given either as a `DocumentArray` or
    a name of the `DocumentArray` that is already pushed to Jina AI Cloud.

    Checks not to upload same dataset twice.

    :param experiment_name: Name of the experiment.
    :param run_name: Name of the run.
    :param train_data: Training data.
    :param eval_data: Evaluation data.
    :param query_data: Query data for `EvaluationCallback`.
    :param index_data: Index data for `EvaluationCallback`.
    :return: Name(s) of the uploaded data.
    """
    _ids2names = dict()
    return (
        push_docarray(
            train_data, f'{DA_PREFIX}-{experiment_name}-{run_name}-train', _ids2names
        ),
        push_docarray(
            eval_data, f'{DA_PREFIX}-{experiment_name}-{run_name}-eval', _ids2names
        ),
        push_docarray(
            query_data, f'{DA_PREFIX}-{experiment_name}-{run_name}-query', _ids2names
        ),
        push_docarray(
            index_data, f'{DA_PREFIX}-{experiment_name}-{run_name}-index', _ids2names
        ),
    )


def push_synthesis_data(
    experiment_name: str,
    run_name: str,
    query_data: Union[str, DocumentArray],
    corpus_data: Union[str, DocumentArray],
) -> Tuple[Optional[str], Optional[str]]:
    """Upload data to Jina AI Cloud and returns their names.

    Uploads all data needed for data synthesis - query data and corpus data.

    Data is given either as a `DocumentArray` or
    a name of the `DocumentArray` that is already pushed to Jina AI Cloud.

    Checks not to upload same dataset twice.

    :param experiment_name: Name of the experiment.
    :param run_name: Name of the run.
    :param query_data: Query data.
    :param corpus_data: Corpus data.
    :return: Names of the uploaded query and corpus data.
    """
    _ids2names = dict()
    return (
        push_docarray(
            query_data, f'{DA_PREFIX}-{experiment_name}-{run_name}-query', _ids2names
        ),
        push_docarray(
            corpus_data, f'{DA_PREFIX}-{experiment_name}-{run_name}-corpus', _ids2names
        ),
    )


def download_artifact(
    client, artifact_id: str, run_name: str, directory: str = ARTIFACTS_DIR
) -> str:
    """Download artifact from Jina AI Cloud by its ID.

    :param client: Hubble client instance.
    :param artifact_id: The artifact id stored in the Jina AI Cloud.
    :param run_name: The name of the run as artifact name to store locally.
    :param directory: Directory where the artifact will be stored.
    :returns: A string that indicates the download path.
    """
    os.makedirs(directory, exist_ok=True)

    path = os.path.join(directory, f'{run_name}.zip')

    return client.hubble_client.download_artifact(
        id=artifact_id, f=path, show_progress=True
    )



from dataclasses import dataclass
from typing import List, Union

from _finetuner.runner.stubs import model
from _finetuner.runner.stubs.model import *  # noqa F401
from _finetuner.runner.stubs.model import _EmbeddingModelStub


def get_header() -> Tuple[str, ...]:
    """Get table header."""
    return 'name', 'task', 'output_dim', 'architecture', 'description'


def get_row(model_stub) -> Tuple[str, ...]:
    """Get table row."""
    return (
        model_stub.display_name,
        model_stub.task,
        str(model_stub.output_shape[1]),
        model_stub.architecture,
        model_stub.description,
    )


def list_model_classes() -> Dict[str, ModelStubType]:
    rv = {}
    members = inspect.getmembers(model, inspect.isclass)
    parent_class = _EmbeddingModelStub
    for name, stub in members:
        if (
            name != 'MLPStub'
            and not name.startswith('_')
            and type(stub) != type
            and issubclass(stub, parent_class)
        ):
            rv[name] = stub
    return rv


@dataclass
class SynthesisModels:
    """Class specifying the models to be used in a data synthesis job.
    :param: relation_miner: The name of the model or list of models to use for
        relation mining.
    :param cross_encoder: The name of the model to use as the cross encoder
    """

    relation_miner: Union[str, List[str]]
    cross_encoder: str


synthesis_model_en = SynthesisModels(
    relation_miner='sbert-base-en',
    cross_encoder='crossencoder-base-en',
)

synthesis_model_multi = SynthesisModels(
    relation_miner='distiluse-base-multi',
    cross_encoder='crossencoder-base-ml',
)
